import pytest
from blessed import Terminal

from pgactivity import views
from pgactivity.config import Flag
from pgactivity.types import UI, QueryMode, SortKey


@pytest.fixture
def term() -> Terminal:
    return Terminal()


@pytest.mark.parametrize(
    "ui, width, expected",
    [
        (
            UI.make(
                query_mode=QueryMode.activities,
                flag=Flag.PID | Flag.DATABASE,
                sort_key=SortKey.cpu,
            ),
            83,
            "PID    DATABASE                     state Query",
        ),
        (
            UI.make(
                query_mode=QueryMode.activities, flag=Flag.CPU, sort_key=SortKey.cpu
            ),
            None,
            "CPU%               state Query",
        ),
        (
            UI.make(
                query_mode=QueryMode.activities, flag=Flag.MEM, sort_key=SortKey.cpu
            ),
            None,
            "MEM%             state Query",
        ),
        (
            UI.make(
                query_mode=QueryMode.blocking,
                flag=Flag.PID
                | Flag.DATABASE
                | Flag.APPNAME
                | Flag.RELATION
                | Flag.CLIENT
                | Flag.WAIT,
                sort_key=SortKey.duration,
            ),
            150,
            "PID    DATABASE                      APP           CLIENT  RELATION          Waiting             state Query",
        ),
    ],
    ids=[
        "pid+database; sort by cpu; activities",
        "cpu; sort by cpu; activities",
        "mem; sort by cpu; activities",
        "many flags; sort by duration; blocking",
    ],
)
def test_columns_header(capsys, term, ui, width, expected):
    views.columns_header(term, ui, width=width)
    out = capsys.readouterr()[0]
    assert out == expected + "\n"



import json
from collections import namedtuple
from unittest.mock import patch

import attr
import pytest

from pgactivity import activities
from pgactivity.types import (
    IOCounter,
    LoadAverage,
    MemoryInfo,
    RunningProcess,
    SwapInfo,
    SystemProcess,
)


@pytest.fixture
def system_processes(datadir):
    with (datadir / "local-processes-input.json").open() as f:
        input_data = json.load(f)

    fs_blocksize = input_data["fs_blocksize"]

    pg_processes = []
    new_system_procs = {}
    system_procs = {}

    running_process_fields = {a.name for a in attr.fields(RunningProcess)}

    def system_process(extras):
        for k in ("io_read", "io_write"):
            try:
                counter = extras.pop(k)
            except KeyError:
                pass
            else:
                extras[k] = IOCounter(counter["count"], counter["bytes"])
        return SystemProcess(**extras)

    for new_proc in input_data["new_processes"].values():
        new_system_procs[new_proc["pid"]] = system_process(new_proc["extras"])
        pg_processes.append(
            RunningProcess(
                **{k: v for k, v in new_proc.items() if k in running_process_fields}
            )
        )

    system_procs = {
        proc["pid"]: system_process(proc["extras"])
        for proc in input_data["processes"].values()
    }

    return pg_processes, system_procs, new_system_procs, fs_blocksize


def test_ps_complete(system_processes):
    pg_processes, system_procs, new_system_procs, fs_blocksize = system_processes

    def sys_get_proc(pid):
        return new_system_procs.pop(pid, None)

    n_system_procs = len(system_procs)

    with patch("pgactivity.activities.sys_get_proc", new=sys_get_proc):
        procs, io_read, io_write = activities.ps_complete(
            pg_processes, system_procs, fs_blocksize
        )

    assert not new_system_procs  # all new system processes consumed

    assert io_read == IOCounter.default()
    assert io_write == IOCounter.default()
    assert len(procs) == len(pg_processes)
    assert len(system_procs) == n_system_procs
    assert {p.pid for p in procs} == {
        6221,
        6222,
        6223,
        6224,
        6225,
        6226,
        6227,
        6228,
        6229,
        6230,
        6231,
        6232,
        6233,
        6234,
        6235,
        6237,
        6238,
        6239,
        6240,
    }


def test_ps_complete_empty_procs(system_processes):
    # same as test_ps_complete() but starting with an empty "system_procs" dict
    pg_processes, __, new_system_procs, fs_blocksize = system_processes

    def sys_get_proc(pid):
        return new_system_procs.pop(pid, None)

    system_procs = {}

    with patch("pgactivity.activities.sys_get_proc", new=sys_get_proc):
        procs, io_read, io_write = activities.ps_complete(
            pg_processes, system_procs, fs_blocksize
        )

    assert not new_system_procs  # all new system processes consumed

    assert io_read == IOCounter.default()
    assert io_write == IOCounter.default()
    assert len(procs) == len(pg_processes)
    assert system_procs


def test_mem_swap_load() -> None:
    pmem = namedtuple("pmem", ["total", "free", "buffers", "cached"])
    vmem = namedtuple("vmem", ["total", "free", "used"])
    with (
        patch(
            "psutil.virtual_memory", return_value=pmem(45, 6, 6, 7)
        ) as virtual_memory,
        patch("psutil.swap_memory", return_value=vmem(8, 6, 2)) as swap_memory,
        patch("os.getloadavg", return_value=(0.14, 0.27, 0.44)) as getloadavg,
    ):
        memory, swap, load = activities.mem_swap_load()
    virtual_memory.assert_called_once_with()
    swap_memory.assert_called_once_with()
    getloadavg.assert_called_once_with()
    assert memory == MemoryInfo(total=45, used=26, free=6, buff_cached=13)
    assert swap == SwapInfo(total=8, used=2, free=6)
    assert load == LoadAverage(0.14, 0.27, 0.44)
    assert memory.pct_used == 57.77777777777778
    assert memory.pct_free == 13.333333333333334
    assert memory.pct_bc == 28.88888888888889
    assert swap.pct_used == 25.0
    assert swap.pct_free == 75.0



from __future__ import annotations

import base64
import functools
import re
import sys
from collections.abc import Iterable, Mapping
from datetime import datetime, timedelta, timezone
from typing import IO, Any

import attr
import humanize

naturalsize = functools.partial(humanize.naturalsize, gnu=True, format="%.2f")
try:
    precisedelta = humanize.precisedelta
except AttributeError:  # humanize < 2.6

    def naturaltimedelta(d: timedelta) -> str:
        """Render a timedelta with seconds truncated.

        >>> d = timedelta(days=5, seconds=15182, microseconds=129198)
        >>> naturaltimedelta(d)
        '5 days, 4:13:00'
        >>> str(d)
        '5 days, 4:13:02.129198'
        """
        d = timedelta(days=d.days, seconds=d.seconds // 60 * 60)
        return str(d)

else:
    naturaltimedelta = functools.partial(
        humanize.precisedelta,
        minimum_unit="minutes",
        format="%.0f",
        suppress=["months", "years"],
    )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class MessagePile:
    """A pile of message.

    >>> p = MessagePile(2)
    >>> p.send("hello")
    >>> p.get()
    'hello'
    >>> p.send("world")
    >>> p.get()
    'world'
    >>> p.get()
    'world'
    >>> p.get()
    """

    n: int
    messages: list[str] = attr.ib(default=attr.Factory(list), init=False)

    def send(self, message: str) -> None:
        self.messages[:] = [message] * self.n

    def get(self) -> str | None:
        if self.messages:
            return self.messages.pop()
        return None


def yn(value: bool) -> str:
    """Return 'Y' or 'N' for a boolean value.

    >>> yn(True)
    'Y'
    >>> yn(False)
    'N'
    """
    return "Y" if value else "N"


def clean_str(string: str) -> str:
    r"""
    Strip and replace some special characters.

    >>> clean_str("\n")
    ''
    >>> clean_str("\n a a  b   b    c \n\t\n c\v\n")
    'a a b b c c'
    """
    msg = str(string)
    msg = msg.replace("\n", " ")
    msg = re.sub(r"\s+", r" ", msg)
    msg = re.sub(r"^\s", r"", msg)
    msg = re.sub(r"\s$", r"", msg)
    return msg


def ellipsis(v: str, width: int) -> str:
    """Shorten a string to specified width with '...' in the middle.

    >>> ellipsis("loooooooooog", 7)
    'lo...og'
    >>> ellipsis("loooooooooog", 6)
    'lo...g'
    >>> ellipsis("short", 10)
    'short'
    """
    lv = len(v)
    if lv <= width:
        return v
    assert width >= 5
    wl = (width - 3) // 2
    return v[: wl + 1] + "..." + v[-wl:]


def get_duration(duration: float | None) -> float:
    """Return 0 if the given duration is negative else, return the duration.

    >>> get_duration(None)
    0
    >>> get_duration(-10)
    0
    >>> get_duration(12)
    12.0
    """
    if duration is None or float(duration) < 0:
        return 0
    return float(duration)


@functools.lru_cache(maxsize=2)
def format_duration(duration: float | None) -> tuple[str, str]:
    """Return a string from 'duration' value along with the color for rendering.

    >>> format_duration(None)
    ('N/A     ', 'green')
    >>> format_duration(-0.000062)
    ('0.000000', 'green')
    >>> format_duration(0.1)
    ('0.100000', 'green')
    >>> format_duration(1.2)
    ('00:01.20', 'yellow')
    >>> format_duration(12345)
    ('205:45.00', 'red')
    >>> format_duration(60001)
    ('16 h', 'red')
    """
    if duration is None:
        return "N/A".ljust(8), "green"

    if duration < 1:
        if duration < 0:
            duration = 0
        ctime = f"{duration:.6f}"
        color = "green"
    elif duration < 60000:
        if duration < 3:
            color = "yellow"
        else:
            color = "red"
        duration_d = timedelta(seconds=float(duration))
        mic = "%.6d" % duration_d.microseconds
        ctime = "{}:{}.{}".format(
            str(duration_d.seconds // 60).zfill(2),
            str(duration_d.seconds % 60).zfill(2),
            mic[:2],
        )
    else:
        ctime = "%s h" % str(int(duration / 3600))
        color = "red"

    return ctime, color


def wait_status(value: None | bool | str) -> str:
    """Display the waiting status of query.

    >>> wait_status(None)
    ''
    >>> wait_status(False)
    'N'
    >>> wait_status(True)
    'Y'
    >>> wait_status("MultiXactTruncationLock")
    'MultiXactTruncationLock'
    """
    if isinstance(value, str):
        return value
    if isinstance(value, bool):
        return yn(value)
    return ""


def short_state(state: str) -> str:
    """Return a short version of query state.

    >>> short_state("active")
    'active'
    >>> short_state("idle in transaction")
    'idle in trans'
    >>> short_state("idle in transaction (aborted)")
    'idle in trans (a)'
    """
    return {
        "idle in transaction": "idle in trans",
        "idle in transaction (aborted)": "idle in trans (a)",
    }.get(state, state)


def osc52_copy(text: str) -> None:
    assert sys.__stderr__ is not None
    buffer = sys.__stderr__.buffer
    buffer.write(b";".join([b"\033]52", b"c", base64.b64encode(text.encode())]) + b"\a")
    buffer.flush()


def csv_write(
    fobj: IO[str],
    procs: Iterable[Mapping[str, Any]],
    *,
    delimiter: str = ";",
) -> None:
    """Store process list into CSV file.

    >>> processes = [
    ...     {'pid': 25199, 'xmin': 1234, 'application_name': '', 'database': 'pgbench', 'user': None,
    ...      'client': 'local', 'cpu': 0.0, 'mem': 0.6504979545924837,
    ...      'read': 0.0, 'write': 0.0, 'state': 'active',
    ...      'query': 'autovacuum: VACUUM ANALYZE public.pgbench_tellers',
    ...      'duration': 0.348789, 'wait': False,
    ...      'io_wait': False, 'is_parallel_worker': False},
    ...     {'pid': 25068, 'xmin': 2345, 'application_name': 'pgbench', 'database': None,
    ...      'user': 'postgres', 'client': 'local', 'cpu': 0.0, 'mem': 2.4694780629380646,
    ...      'read': 278536.76590087387, 'write': 835610.2977026217,
    ...      'state': 'idle in transaction',
    ...      'query': 'INSERT INTO pgbench_history (tid, bid, aid, delta, mtime) VALUES (625, 87, 4368910, -341, CURRENT_TIMESTAMP);',
    ...      'duration': 0.000105, 'wait': False, 'io_wait': False,
    ...      'is_parallel_worker': False},
    ...     {'pid': 25379, 'xmin': 3456, 'application_name': 'pgbench', 'database': 'pgbench',
    ...      'user': 'postgres', 'client': 'local', 'state': 'active',
    ...      'query': 'UPDATE pgbench_branches SET bbalance = bbalance + -49 WHERE bid = 73;',
    ...      'duration': 0, 'wait': False},
    ...     {'pid': 25392, 'xmin': 4567, 'application_name': 'pgbench', 'database': 'pgbench',
    ...      'user': 'postgres', 'client': 'local', 'state': 'active',
    ...      'query': 'BEGIN;', 'duration': 0, 'wait': False}
    ... ]
    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile(mode='w+') as f:
    ...     csv_write(f, processes[:2])
    ...     csv_write(f, processes[2:])
    ...     _ = f.seek(0)
    ...     content = f.read()
    >>> print(content, end="")  # doctest: +ELLIPSIS
    datetimeutc;pid;xmin;database;appname;user;client;cpu;memory;read;write;duration;wait;io_wait;state;query
    "...-...-...T...Z";"25199";"1234";"pgbench";"";"None";"local";"0.0";"0.6504979545924837";"0.0";"0.0";"0.348789";"N";"N";"active";"autovacuum: VACUUM ANALYZE public.pgbench_tellers"
    "...-...-...T...Z";"25068";"2345";"";"pgbench";"postgres";"local";"0.0";"2.4694780629380646";"278536.76590087387";"835610.2977026217";"0.000105";"N";"N";"idle in transaction";"INSERT INTO pgbench_history (tid, bid, aid, delta, mtime) VALUES (625, 87, 4368910, -341, CURRENT_TIMESTAMP);"
    "...-...-...T...Z";"25379";"3456";"pgbench";"pgbench";"postgres";"local";"N/A";"N/A";"N/A";"N/A";"0";"N";"N/A";"active";"UPDATE pgbench_branches SET bbalance = bbalance + -49 WHERE bid = 73;"
    "...-...-...T...Z";"25392";"4567";"pgbench";"pgbench";"postgres";"local";"N/A";"N/A";"N/A";"N/A";"0";"N";"N/A";"active";"BEGIN;"
    """

    def clean_str_csv(s: str) -> str:
        return clean_str(s).replace('"', '\\"')

    if fobj.tell() == 0:
        # First line then write CSV header
        fobj.write(
            delimiter.join(
                [
                    "datetimeutc",
                    "pid",
                    "xmin",
                    "database",
                    "appname",
                    "user",
                    "client",
                    "cpu",
                    "memory",
                    "read",
                    "write",
                    "duration",
                    "wait",
                    "io_wait",
                    "state",
                    "query",
                ]
            )
            + "\n"
        )

    def yn_na(value: bool | None) -> str:
        if value is None:
            return "N/A"
        return yn(value)

    for p in procs:
        dt = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ")
        pid = p.get("pid", "N/A")
        xmin = p.get("xmin", "N/A")
        database = p.get("database", "N/A") or ""
        appname = p.get("application_name", "N/A")
        user = p.get("user", "N/A")
        client = p.get("client", "N/A")
        cpu = p.get("cpu", "N/A")
        mem = p.get("mem", "N/A")
        read = p.get("read", "N/A")
        write = p.get("write", "N/A")
        duration = p.get("duration", "N/A")
        wait = yn_na(p.get("wait"))
        io_wait = yn_na(p.get("io_wait"))
        state = p.get("state", "N/A")
        query = clean_str_csv(p.get("query", "N/A"))
        fobj.write(
            delimiter.join(
                [
                    f'"{dt}"',
                    f'"{pid}"',
                    f'"{xmin}"',
                    f'"{database}"',
                    f'"{appname}"',
                    f'"{user}"',
                    f'"{client}"',
                    f'"{cpu}"',
                    f'"{mem}"',
                    f'"{read}"',
                    f'"{write}"',
                    f'"{duration}"',
                    f'"{wait}"',
                    f'"{io_wait}"',
                    f'"{state}"',
                    f'"{query}"',
                ]
            )
            + "\n"
        )



from __future__ import annotations

import logging
import pathlib
import threading
from typing import Any

import psycopg
import psycopg.errors
import pytest
from psycopg import sql
from psycopg.conninfo import make_conninfo

from pgactivity import pg

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


def pytest_report_header(config: Any) -> list[str]:
    return [f"psycopg: {pg.__version__}"]


@pytest.fixture(scope="session")
def datadir() -> pathlib.Path:
    return pathlib.Path(__file__).parent / "data"


@pytest.fixture
def database_factory(postgresql):
    dbnames = set()

    def createdb(dbname: str, encoding: str, locale: str | None = None) -> None:
        with psycopg.connect(postgresql.info.dsn, autocommit=True) as conn:
            qs = sql.SQL(
                "CREATE DATABASE {dbname} ENCODING {encoding} TEMPLATE template0"
            ).format(dbname=sql.Identifier(dbname), encoding=sql.Identifier(encoding))
            if locale:
                qs = sql.SQL(" ").join(
                    [
                        qs,
                        sql.SQL("LOCALE {locale}").format(
                            locale=sql.Identifier(locale)
                        ),
                    ]
                )
            conn.execute(qs)
        dbnames.add(dbname)

    yield createdb

    with psycopg.connect(postgresql.info.dsn, autocommit=True) as conn:
        for dbname in dbnames:
            conn.execute(
                sql.SQL("DROP DATABASE IF EXISTS {dbname} WITH (FORCE)").format(
                    dbname=sql.Identifier(dbname)
                )
            )


@pytest.fixture
def execute(postgresql):
    """Create a thread and return an execute() function that will run SQL queries in that
    thread.
    """
    threads_and_cnx = []

    def execute(
        query: str,
        commit: bool = False,
        autocommit: bool = False,
        dbname: str | None = None,
    ) -> None:
        dsn, kwargs = postgresql.info.dsn, {}
        if dbname:
            kwargs["dbname"] = dbname
        conn = psycopg.connect(make_conninfo(dsn, **kwargs))
        conn.autocommit = autocommit

        def _execute() -> None:
            LOGGER.info(
                "running query %s (commit=%s, autocommit=%s) using connection <%s>",
                query,
                commit,
                autocommit,
                id(conn),
            )
            with conn.cursor() as c:
                try:
                    c.execute(query)
                except (
                    psycopg.errors.AdminShutdown,
                    psycopg.errors.QueryCanceled,
                ):
                    return
                if not autocommit and commit:
                    conn.commit()
            LOGGER.info("query %s finished", query)

        thread = threading.Thread(target=_execute, daemon=True)
        thread.start()
        threads_and_cnx.append((thread, conn))

    yield execute

    for thread, conn in threads_and_cnx:
        thread.join(timeout=2)
        LOGGER.info("closing connection <%s>", id(conn))
        conn.close()



from __future__ import annotations

from pathlib import Path
from typing import Any

import attr
import pytest

from pgactivity.config import Configuration, ConfigurationError, Flag, UISection


def test_flag() -> None:
    f = Flag(3)
    assert f == Flag.APPNAME | Flag.DATABASE
    assert f | Flag.CLIENT == Flag.CLIENT | Flag.APPNAME | Flag.DATABASE
    f ^= Flag.APPNAME
    assert f == Flag.DATABASE


def test_flag_load() -> None:
    options = {
        "appname": True,
        "client": True,
        "cpu": True,
        "database": True,
        "mem": True,
        "pid": None,
        "read": True,
        "relation": True,
        "time": True,
        "user": True,
        "wait": None,
        "write": True,
        "xmin": True,
    }
    flag = Flag.load(None, is_local=True, **options)
    assert (
        flag
        == Flag.PID
        | Flag.IOWAIT
        | Flag.MODE
        | Flag.TYPE
        | Flag.RELATION
        | Flag.WAIT
        | Flag.TIME
        | Flag.WRITE
        | Flag.READ
        | Flag.MEM
        | Flag.CPU
        | Flag.USER
        | Flag.CLIENT
        | Flag.APPNAME
        | Flag.DATABASE
        | Flag.XMIN
    )
    cfg = Configuration(
        name="test",
        values=dict(
            pid=UISection(hidden=True),
            relation=UISection(hidden=False),
            wait=UISection(hidden=False),
        ),
    )
    flag = Flag.load(cfg, is_local=False, **options)
    assert (
        flag
        == Flag.MODE
        | Flag.TYPE
        | Flag.RELATION
        | Flag.WAIT
        | Flag.TIME
        | Flag.USER
        | Flag.CLIENT
        | Flag.APPNAME
        | Flag.DATABASE
        | Flag.XMIN
    )
    options["database"] = False
    options["time"] = False
    options["pid"] = False
    cfg = Configuration(
        name="test",
        values=dict(database=UISection(hidden=False), relation=UISection(hidden=True)),
    )
    flag = Flag.load(cfg, is_local=False, **options)
    assert (
        flag
        == Flag.MODE
        | Flag.TYPE
        | Flag.WAIT
        | Flag.USER
        | Flag.CLIENT
        | Flag.APPNAME
        | Flag.XMIN
    )


def asdict(cfg: Configuration) -> dict[str, Any]:
    return {k: attr.asdict(v) for k, v in cfg.items()}


def test_error() -> None:
    cfg = Configuration(name="test", values={})
    with pytest.raises(ConfigurationError, match="test error") as cm:
        raise cfg.error("test error")
    assert cm.value.filename == "test"


def test_lookup(tmp_path: Path) -> None:
    cfg = Configuration.lookup(None, user_config_home=tmp_path)
    assert cfg is None

    (tmp_path / "pg_activity.conf").write_text(
        "\n".join(
            [
                "[client]",
                "width=5",
                "color=cyan",
                "[header]",
                "show_instance=no",
            ]
        )
    )
    cfg = Configuration.lookup(None, user_config_home=tmp_path)
    assert cfg is not None and asdict(cfg) == {
        "client": {"hidden": False, "width": 5, "color": "cyan"},
        "header": {"show_instance": False, "show_system": True, "show_workers": True},
    }

    (tmp_path / "pg_activity").mkdir()
    (tmp_path / "pg_activity" / "x.conf").write_text(
        "\n".join(
            ["[database]", "hidden= on", "width = 3 ", "[header]", "show_workers=no"]
        )
    )
    cfg = Configuration.lookup("x", user_config_home=tmp_path)
    assert cfg is not None and asdict(cfg) == {
        "database": {"hidden": True, "width": 3, "color": None},
        "header": {"show_instance": True, "show_system": True, "show_workers": False},
    }

    with pytest.raises(FileNotFoundError):
        Configuration.lookup("y", user_config_home=tmp_path)


no_header = {
    "header": {k: False for k in ("show_instance", "show_system", "show_workers")}
}
columns = (
    "database",
    "user",
    "client",
    "cpu",
    "mem",
    "read",
    "write",
    "appname",
    "xmin",
)
narrow = {k: {"hidden": True, "width": None, "color": None} for k in columns}
wide = {k: {"hidden": False, "width": None, "color": None} for k in columns}
minimal = {**no_header, **narrow}


@pytest.mark.parametrize(
    "profile, expected",
    [
        ("minimal", minimal),
        ("narrow", narrow),
        ("wide", wide),
    ],
)
def test_lookup_builtin_profiles(
    tmp_path: Path, profile: str, expected: dict[str, Any]
) -> None:
    cfg = Configuration.lookup(profile, user_config_home=tmp_path)
    assert cfg is not None and asdict(cfg) == expected



from __future__ import annotations

import time

import attr
import psycopg
import pytest
from psycopg.errors import WrongObjectType

from pgactivity import types
from pgactivity.data import Data


def retry(fct, msg: str, timeout: int = 2):
    count = int(timeout / 0.1)
    for _ in range(count):
        time.sleep(0.1)
        value = fct()
        if value:
            return value
    pytest.fail(msg)


@pytest.fixture
def data(postgresql):
    obj = Data.pg_connect(
        host=postgresql.info.host,
        port=postgresql.info.port,
        database=postgresql.info.dbname,
        user=postgresql.info.user,
    )
    yield obj
    obj.pg_conn.close()


def test_pg_is_local(postgresql, data):
    assert data.pg_is_local()


def test_pg_is_local_access(postgresql, data):
    assert data.pg_is_local_access()


def test_pg_get_server_information(data):
    data.pg_get_server_information(None)


def test_activities(postgresql, data):
    postgresql.execute("SELECT pg_sleep(1)")
    (running,) = data.pg_get_activities()
    assert "pg_sleep" in running.query
    assert running.state == "idle in transaction"
    if postgresql.info.server_version >= 100000:
        assert running.wait == "ClientRead"
    else:
        assert running.wait is None
    assert not running.is_parallel_worker


def test_blocking_waiting(postgresql, data, execute):
    postgresql.execute("CREATE TABLE t AS (SELECT 'init'::text s)")
    postgresql.commit()
    execute("UPDATE t SET s = 'blocking'")
    execute("UPDATE t SET s = 'waiting 1'", commit=True)
    execute("UPDATE t SET s = 'waiting 2'", commit=True)
    blocking = retry(data.pg_get_blocking, msg="could not fetch blocking queries")
    waiting = data.pg_get_waiting()
    assert len(blocking) == 2
    assert len(waiting) == 2
    assert "blocking" in blocking[0].query
    assert "waiting 1" in waiting[0].query and "waiting 2" in waiting[1].query
    if postgresql.info.server_version >= 100000:
        assert blocking[0].wait == "ClientRead"
        assert blocking[1].wait == "transactionid"
    assert str(blocking[0].type) == "transactionid"
    assert str(blocking[1].type) == "tuple"


def test_pg_get_blocking_virtualxid(postgresql, data, execute):
    postgresql.execute("CREATE TABLE t AS (SELECT 'init'::text s)")
    postgresql.commit()
    execute("UPDATE t SET s = 'blocking'")
    execute("CREATE INDEX CONCURRENTLY ON t(s)", autocommit=True)
    (blocking,) = retry(data.pg_get_blocking, msg="could not fetch blocking queries")
    (waiting,) = data.pg_get_waiting()
    assert "blocking" in blocking.query
    assert "CREATE INDEX CONCURRENTLY ON t(s)" in waiting.query
    assert str(blocking.type) == "virtualxid"


def test_cancel_backend(postgresql, data):
    postgresql.execute("SELECT pg_sleep(1)")
    (running,) = data.pg_get_activities()
    assert data.pg_cancel_backend(running.pid)


def test_terminate_backend(postgresql, data):
    postgresql.execute("SELECT pg_sleep(1)")
    (running,) = data.pg_get_activities()
    assert data.pg_terminate_backend(running.pid)
    retry(lambda: not data.pg_get_activities(), msg="some processes are still active")


def test_encoding(postgresql, data, execute):
    """Test for issue #149, #332."""
    conninfo = postgresql.info.dsn
    # plateform specific locales (Centos, Ubuntu)
    encodings = ["fr_FR.latin1", "fr_FR.88591", "fr_FR.8859-1"]
    with psycopg.connect(conninfo, autocommit=True) as conn:
        for encoding in encodings:
            try:
                conn.execute(
                    f"CREATE DATABASE latin1 ENCODING 'latin1' TEMPLATE template0 LC_COLLATE '{encoding}' LC_CTYPE '{encoding}'"
                )
            except WrongObjectType:
                continue
            else:
                break
        else:
            pytest.skip(
                f"could not create a database with encoding amongst {', '.join(encodings)}"
            )

    with psycopg.connect(conninfo, dbname="latin1") as conn:
        conn.execute("CREATE TABLE tbl AS (SELECT 'initilialized éléphant' s)")
        conn.commit()

    execute("UPDATE tbl SET s = 'blocking éléphant'", dbname="latin1")
    execute("UPDATE tbl SET s = 'waiting éléphant'", dbname="latin1", commit=True)
    running = retry(data.pg_get_activities, msg="could not fetch activities")
    assert "blocking éléphant" in running[0].query
    (waiting,) = retry(data.pg_get_waiting, "no waiting process")
    assert waiting.query and "waiting éléphant" in waiting.query
    (blocking,) = data.pg_get_blocking()
    assert blocking.query and "blocking éléphant" in blocking.query


@pytest.mark.parametrize("encoding", ["utf-8", "latin1", "sql_ascii"])
def test_client_encoding(postgresql, encoding: str) -> None:
    data = Data.pg_connect(
        host=postgresql.info.host,
        port=postgresql.info.port,
        database=postgresql.info.dbname,
        user=postgresql.info.user,
        dsn=f"client_encoding={encoding}",
    )
    try:
        assert data.pg_version.startswith(
            f"PostgreSQL {str(postgresql.info.server_version)[:2]}"
        )
    finally:
        data.pg_conn.close()


@pytest.mark.parametrize(
    "pyenc, pgenc, locale",
    [
        ("utf8", "UTF8", None),
        ("ascii", "SQL_ASCII", None),
        ("unknown", "EUC_TW", "zh_TW.euctw"),
    ],
)
def test_postgres_and_python_encoding(
    database_factory, pyenc: str, pgenc: str, locale: str | None, data, postgresql
) -> None:
    dbname = pyenc
    try:
        database_factory(dbname, encoding=pgenc, locale=locale)
    except WrongObjectType:
        pytest.skip(f"could not create a database with encoding '{pgenc}'")
    with psycopg.connect(
        postgresql.info.dsn, dbname=dbname, client_encoding="utf-8"
    ) as conn:
        conn.execute("SELECT 'encoding'")
        (running,) = data.pg_get_activities()
    assert "'encoding'" in running.query
    assert running.encoding == pgenc
    assert running.database == dbname


def test_filters_dbname(data, execute):
    data_filtered = attr.evolve(data, filters=types.Filters(dbname="temp"))
    execute("SELECT pg_sleep(2)", dbname="template1", autocommit=True)
    nbconn = retry(
        data.pg_get_server_information,
        msg="could not get active connections for filtered DBNAME",
    )
    nbconn_filtered = retry(
        data_filtered.pg_get_server_information,
        msg="could not get active connections for filtered DBNAME",
    )
    assert nbconn.active_connections == 2
    assert nbconn_filtered.active_connections == 1



from __future__ import annotations

import configparser
import enum
import importlib.resources
import io
import os
from collections.abc import ItemsView
from pathlib import Path
from typing import IO, Any, TypeVar, Union

import attr
from attr import validators

from .compat import gt


def read_resource(pkgname: str, dirname: str, *args: str) -> str | None:
    resource = importlib.resources.files(pkgname).joinpath(dirname)
    for arg in args:
        resource = resource.joinpath(arg)
    if resource.is_file():
        return resource.read_text()
    return None


class ConfigurationError(Exception):
    def __init__(self, filename: str, *args: Any) -> None:
        super().__init__(*args)
        self.filename = filename

    @property
    def message(self) -> str:
        return super().__str__()

    def __str__(self) -> str:
        return f"invalid configuration '{self.filename}': {self.message}"


class InvalidSection(ConfigurationError):
    def __init__(self, section: str, *args: Any) -> None:
        super().__init__(*args)
        self.section = section

    @property
    def message(self) -> str:
        return f"invalid section '{self.section}'"


class InvalidOptions(ConfigurationError):
    def __init__(self, section: str, message: str, *args: Any) -> None:
        super().__init__(*args)
        self.section = section
        self._message = message

    @property
    def message(self) -> str:
        return f"invalid option(s) in '{self.section}': {self._message}"


class Flag(enum.Flag):
    """Column flag.

    >>> Flag.names()
    ['database', 'appname', 'client', 'user', 'cpu', 'mem', 'read', 'write', 'time', 'wait', 'relation', 'type', 'mode', 'iowait', 'pid', 'xmin']
    >>> Flag.all()  # doctest: +ELLIPSIS
    <Flag...: 65535>
    """

    DATABASE = enum.auto()
    APPNAME = enum.auto()
    CLIENT = enum.auto()
    USER = enum.auto()
    CPU = enum.auto()
    MEM = enum.auto()
    READ = enum.auto()
    WRITE = enum.auto()
    TIME = enum.auto()
    WAIT = enum.auto()
    RELATION = enum.auto()
    TYPE = enum.auto()
    MODE = enum.auto()
    IOWAIT = enum.auto()
    PID = enum.auto()
    XMIN = enum.auto()

    @classmethod
    def names(cls) -> list[str]:
        rv = []
        for f in cls:
            assert f.name
            rv.append(f.name.lower())
        return rv

    @classmethod
    def all(cls) -> Flag:
        value = cls(0)
        for f in cls:
            value |= f
        return value

    @classmethod
    def from_config(cls, config: Configuration) -> Flag:
        value = cls(0)
        for f in cls:
            assert f.name is not None
            try:
                cfg = config[f.name.lower()]
            except KeyError:
                pass
            else:
                assert isinstance(cfg, UISection)
                if cfg.hidden:
                    continue
            value |= f
        return value

    @classmethod
    def load(
        cls,
        config: Configuration | None,
        *,
        is_local: bool,
        appname: bool | None,
        client: bool | None,
        cpu: bool | None,
        database: bool | None,
        mem: bool | None,
        pid: bool | None,
        read: bool | None,
        time: bool | None,
        user: bool | None,
        wait: bool | None,
        write: bool | None,
        xmin: bool | None,
        **kwargs: Any,
    ) -> Flag:
        """Build a Flag value from command line options."""
        if config:
            flag = cls.from_config(config)
        else:
            flag = cls.all()
        for opt, value in (
            (appname, cls.APPNAME),
            (client, cls.CLIENT),
            (cpu, cls.CPU),
            (database, cls.DATABASE),
            (mem, cls.MEM),
            (pid, cls.PID),
            (read, cls.READ),
            (time, cls.TIME),
            (user, cls.USER),
            (wait, cls.WAIT),
            (write, cls.WRITE),
            (xmin, cls.XMIN),
        ):
            if opt is True:
                flag |= value
            elif opt is False:
                flag ^= value
        # Remove some if no running against local pg server.
        if not is_local and (flag & cls.CPU):
            flag ^= cls.CPU
        if not is_local and (flag & cls.MEM):
            flag ^= cls.MEM
        if not is_local and (flag & cls.READ):
            flag ^= cls.READ
        if not is_local and (flag & cls.WRITE):
            flag ^= cls.WRITE
        if not is_local and (flag & cls.IOWAIT):
            flag ^= cls.IOWAIT
        return flag


class BaseSectionMixin:
    @classmethod
    def check_options(
        cls: type[attr.AttrsInstance], section: configparser.SectionProxy
    ) -> list[str]:
        """Check that items of 'section' conform to known attributes of this class and
        return the list of know options.
        """
        known_options = {f.name for f in attr.fields(cls)}
        unknown_options = set(section) - set(known_options)
        if unknown_options:
            raise ValueError(f"invalid option(s): {', '.join(sorted(unknown_options))}")
        return list(sorted(known_options))


@attr.s(auto_attribs=True, frozen=True, slots=True)
class HeaderSection(BaseSectionMixin):
    show_instance: bool = True
    show_system: bool = True
    show_workers: bool = True

    _T = TypeVar("_T", bound="HeaderSection")

    @classmethod
    def from_config_section(cls: type[_T], section: configparser.SectionProxy) -> _T:
        values: dict[str, bool] = {}
        for optname in cls.check_options(section):
            value = section.getboolean(optname)
            if value is not None:
                values[optname] = value
        return cls(**values)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class UISection(BaseSectionMixin):
    hidden: bool = False
    width: int | None = attr.ib(default=None, validator=validators.optional(gt(0)))
    color: str | None = attr.ib(default=None)

    _T = TypeVar("_T", bound="UISection")

    @classmethod
    def from_config_section(cls: type[_T], section: configparser.SectionProxy) -> _T:
        cls.check_options(section)
        values: dict[str, Any] = {}
        hidden = section.getboolean("hidden")
        if hidden is not None:
            values["hidden"] = hidden
        values["width"] = section.getint("width")
        values["color"] = section.get("color")
        return cls(**values)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class BuiltinProfile:
    name: str
    content: IO[str]

    @classmethod
    def get(cls, name: str) -> BuiltinProfile | None:
        content = read_resource("pgactivity", "profiles", f"{name}.conf")
        if content is not None:
            return cls(name, io.StringIO(content))
        return None


USER_CONFIG_HOME = Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config"))
ETC = Path("/etc")


Value = Union[HeaderSection, UISection]


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Configuration:
    name: str
    values: dict[str, Value]

    def __getitem__(self, name: str) -> Value:
        return self.values.__getitem__(name)

    def get(self, name: str, default: Value | None = None) -> Value | None:
        return self.values.get(name, default)

    def items(self) -> ItemsView[str, Value]:
        return self.values.items()

    def header(self) -> HeaderSection | None:
        return self.get("header")  # type: ignore[return-value]

    _T = TypeVar("_T", bound="Configuration")

    @classmethod
    def parse(cls: type[_T], f: IO[str], name: str) -> _T:
        r"""Parse configuration from 'f'.

        >>> from io import StringIO
        >>> from pprint import pprint

        >>> f = StringIO('[header]\nshow_workers=false\n[client]\nhidden=true\ncolor=green\n')
        >>> cfg = Configuration.parse(f, "f.ini")
        >>> cfg.name
        'f.ini'
        >>> pprint(cfg.values)
        {'client': UISection(hidden=True, width=None, color='green'),
         'header': HeaderSection(show_instance=True, show_system=True, show_workers=False)}

        >>> bad = StringIO("[global]\nx=1")
        >>> Configuration.parse(bad, "bad.ini")
        Traceback (most recent call last):
          ...
        pgactivity.config.InvalidSection: invalid configuration 'bad.ini': invalid section 'global'
        >>> bad = StringIO("[xxx]\n")
        >>> Configuration.parse(bad, "bad.ini")
        Traceback (most recent call last):
          ...
        pgactivity.config.InvalidSection: invalid configuration 'bad.ini': invalid section 'xxx'
        >>> bad = StringIO("[cpu]\nx=1")
        >>> Configuration.parse(bad, "bad.ini")
        Traceback (most recent call last):
          ...
        pgactivity.config.InvalidOptions: invalid configuration 'bad.ini': invalid option(s) in 'cpu': invalid option(s): x
        >>> bad = StringIO("[mem]\nwidth=-2")
        >>> Configuration.parse(bad, "bad.ini")
        Traceback (most recent call last):
          ...
        pgactivity.config.InvalidOptions: invalid configuration 'bad.ini': invalid option(s) in 'mem': 'width' must be > 0: -2
        >>> bad = StringIO("[mem]\nwidth=xyz")
        >>> Configuration.parse(bad, "bad.ini")
        Traceback (most recent call last):
          ...
        pgactivity.config.InvalidOptions: invalid configuration 'bad.ini': invalid option(s) in 'mem': invalid literal for int() with base 10: 'xyz'
        >>> bad = StringIO("not some INI??")
        >>> Configuration.parse(bad, "bad.txt")
        Traceback (most recent call last):
          ...
        pgactivity.config.ConfigurationError: invalid configuration 'bad.txt': failed to parse INI: File contains no section headers.
        file: '<???>', line: 1
        'not some INI??'
        """
        p = configparser.ConfigParser(default_section="global", strict=True)
        try:
            p.read_file(f)
        except configparser.Error as e:
            raise ConfigurationError(name, f"failed to parse INI: {e}") from None
        known_sections = set(Flag.names())
        config: dict[str, HeaderSection | UISection] = {}
        for sname, section in p.items():
            if sname == p.default_section:
                if section:
                    raise InvalidSection(p.default_section, name)
                continue
            if sname == "header":
                config[sname] = HeaderSection.from_config_section(section)
                continue
            if sname not in known_sections:
                raise InvalidSection(sname, name)
            try:
                config[sname] = UISection.from_config_section(section)
            except ValueError as e:
                raise InvalidOptions(sname, str(e), name) from None
        return cls(name=name, values=config)

    @classmethod
    def lookup(
        cls: type[_T],
        profile: str | None,
        *,
        user_config_home: Path = USER_CONFIG_HOME,
        etc: Path = ETC,
    ) -> _T | None:
        if profile is None:
            for base in (user_config_home, etc):
                fpath = base / "pg_activity.conf"
                if fpath.exists():
                    with fpath.open() as f:
                        return cls.parse(f, str(fpath))
            return None

        assert profile  # per argument validation
        fname = f"{profile}.conf"
        bases = (user_config_home / "pg_activity", etc / "pg_activity")
        for base in bases:
            fpath = base / fname
            if fpath.exists():
                with fpath.open() as f:
                    return cls.parse(f, str(fpath))

        builtin_profile = BuiltinProfile.get(profile)
        if builtin_profile is not None:
            return cls.parse(builtin_profile.content, builtin_profile.name)

        raise FileNotFoundError(f"profile {profile!r} not found")

    def error(self, message: str) -> ConfigurationError:
        return ConfigurationError(self.name, message)



import functools
import pathlib

here = pathlib.Path(__file__).parent


@functools.cache
def get(name: str) -> str:
    r"""Return the SQL query contained in named file, ignoring commented lines.

    >>> get('get_version')
    'SELECT version() AS pg_version;'
    >>> print(get('get_pg_activity_post_130000')[:101])
    SELECT
          a.pid AS pid,
          a.backend_xmin AS xmin,
          a.application_name AS application_name
    """
    path = here / f"{name}.sql"
    s = "-- "
    with path.open() as f:
        return "\n".join(
            line.rstrip().split(s, 1)[0]
            for line in f
            if line.strip() and not line.startswith(s)
        )



from __future__ import annotations

from blessed.keyboard import Keystroke

from . import keys
from .config import Flag
from .types import DurationMode, QueryMode, SortKey, enum_next


def refresh_time(
    key: str | None, value: float, minimum: float = 0.5, maximum: float = 5
) -> float:
    """Return an updated refresh time interval from input key respecting bounds.

    >>> refresh_time("+", 1)
    2
    >>> refresh_time("+", 5)
    5
    >>> refresh_time("+", 5, maximum=10)
    6
    >>> refresh_time("-", 2)
    1
    >>> refresh_time("-", 1)
    0.5
    >>> refresh_time("-", 0.5)
    0.5
    >>> refresh_time("=", 42)
    Traceback (most recent call last):
        ...
    ValueError: =
    """
    if key == keys.REFRESH_TIME_DECREASE:
        return max(value - 1, minimum)
    elif key == keys.REFRESH_TIME_INCREASE:
        return min(int(value + 1), maximum)
    raise ValueError(key)


def duration_mode(key: Keystroke, mode: DurationMode) -> DurationMode:
    """Return the updated duration mode matching input key.

    >>> from blessed.keyboard import Keystroke as k

    >>> duration_mode(k("42"), DurationMode.query).name
    'query'
    >>> duration_mode(k("T"), DurationMode.transaction).name
    'backend'
    """
    if key == keys.CHANGE_DURATION_MODE:
        return enum_next(mode)
    return mode


def wrap_query(key: Keystroke, wrap: bool) -> bool:
    """Return the updated 'wrap' value depending on input key.

    >>> from blessed.keyboard import Keystroke as k

    >>> wrap_query(k("42"), True)
    True
    >>> wrap_query(k("v"), False)
    True
    """
    if key == keys.WRAP_QUERY:
        return not wrap
    return wrap


def query_mode(key: Keystroke) -> QueryMode | None:
    """Return the query mode matching input key or None.

    >>> import curses
    >>> from blessed.keyboard import Keystroke as k

    >>> query_mode(k("42"))
    >>> query_mode(k("1")).name
    'activities'
    >>> query_mode(k(code=curses.KEY_F3)).name
    'blocking'
    """
    if key.is_sequence and key.code:
        try:
            return keys.QUERYMODE_FROM_KEYS[key.code]
        except KeyError:
            pass
    return keys.QUERYMODE_FROM_KEYS.get(key)


def sort_key_for(key: Keystroke, query_mode: QueryMode, flag: Flag) -> SortKey | None:
    """Return the sort key matching input key or None.

    >>> from blessed.keyboard import Keystroke as k
    >>> from pgactivity.types import QueryMode

    >>> flag = Flag.all()

    Unhandled key:
    >>> sort_key_for(k("1"), QueryMode.activities, flag)

    In activities mode, 'm', 'w', 't', ... keys are handled:
    >>> sort_key_for(k("m"), QueryMode.activities, flag).name
    'mem'
    >>> sort_key_for(k("w"), QueryMode.activities, flag).name
    'write'
    >>> sort_key_for(k("t"), QueryMode.activities, flag).name
    'duration'
    >>> sort_key_for(k("c"), QueryMode.activities, flag).name
    'cpu'

    In other modes, the default sort key is always returned:
    >>> sort_key_for(k("m"), QueryMode.waiting, flag).name
    'duration'

    When flag does not match given sort key, return None:
    >>> flag ^= Flag.CPU
    >>> sort_key_for(k("c"), QueryMode.activities, flag)
    >>> sort_key_for(k("m"), QueryMode.activities, flag).name
    'mem'
    >>> flag ^= Flag.MEM
    >>> sort_key_for(k("m"), QueryMode.activities, flag)
    """
    if query_mode != QueryMode.activities:
        return SortKey.default()
    try:
        sort_key, required_flag = {
            keys.SORTBY_CPU: (SortKey.cpu, Flag.CPU),
            keys.SORTBY_MEM: (SortKey.mem, Flag.MEM),
            keys.SORTBY_READ: (SortKey.read, Flag.READ),
            keys.SORTBY_TIME: (SortKey.duration, Flag.TIME),
            keys.SORTBY_WRITE: (SortKey.write, Flag.WRITE),
        }[key]
    except KeyError:
        return None
    if flag & required_flag:
        return sort_key
    return None



from __future__ import annotations

import operator
from importlib.metadata import version
from typing import Any

import attr
import attr.validators
import blessed

ATTR_VERSION = tuple(int(x) for x in version("attrs").split(".", 2)[:2])
BLESSED_VERSION = tuple(int(x) for x in version("blessed").split(".", 2)[:2])

if ATTR_VERSION < (18, 1):

    def fields_dict(cls: Any) -> dict[str, Any]:
        return {a.name: a for a in cls.__attrs_attrs__}

else:
    fields_dict = attr.fields_dict

if ATTR_VERSION < (21, 3):

    @attr.s(auto_attribs=True, frozen=True, slots=True)
    class gt:
        bound: int

        def __call__(self, instance: Any, attribute: Any, value: Any) -> None:
            if not operator.gt(value, self.bound):
                raise ValueError(f"'{attribute.name}' must be > {self.bound}: {value}")

else:
    gt = attr.validators.gt  # type: ignore[assignment,misc]


if BLESSED_VERSION < (1, 17):

    def link(term: blessed.Terminal, url: str, text: str, url_id: str = "") -> str:
        return url

else:

    def link(term: blessed.Terminal, url: str, text: str, url_id: str = "") -> str:
        return term.link(url, text, url_id=url_id)



from pgactivity import cli


def test_parser() -> None:
    parser = cli.get_parser()
    ns = parser.parse_args(
        ["--no-db-size", "-w", "-p", "5433", "--no-pid", "--no-app-name"]
    )
    assert vars(ns) == {
        "profile": None,
        "blocksize": 4096,
        "rds": False,
        "output": None,
        "dbsize": False,
        "tempfiles": None,
        "walreceiver": None,
        "wrap_query": True,
        "durationmode": "1",
        "minduration": 0,
        "filters": [],
        "debug_file": None,
        "help": False,
        "connection_string": "",
        "host": None,
        "port": "5433",
        "username": None,
        "dbname": None,
        "pid": False,
        "database": None,
        "user": None,
        "client": None,
        "cpu": None,
        "mem": None,
        "read": None,
        "write": None,
        "time": None,
        "wait": None,
        "xmin": None,
        "appname": False,
        "header_show_instance": None,
        "header_show_system": None,
        "header_show_workers": None,
        "hide_queries_in_logs": False,
        "refresh": 2,
    }


def test_parser_flag_on() -> None:
    parser = cli.get_parser()
    ns = parser.parse_args(["--pid", "--no-app-name"])
    assert ns.pid is True
    assert ns.appname is False
    assert ns.wait is None



from __future__ import annotations

from . import utils

PINNED_COLOR = "bold_yellow"
FOCUSED_COLOR = "cyan_reverse"


def short_state(state: str) -> str | None:
    state = utils.short_state(state)
    if state == "active":
        return "green"
    elif state == "idle in trans":
        return "yellow"
    elif state == "idle in trans (a)":
        return "red"
    return None


def lock_mode(mode: str) -> str:
    if mode in (
        "ExclusiveLock",
        "RowExclusiveLock",
        "AccessExclusiveLock",
    ):
        return "bold_red"
    else:
        return "bold_yellow"


def wait(value: bool) -> str:
    return "red" if value else "green"



from __future__ import annotations

import curses
from typing import Any

import attr
from blessed.keyboard import Keystroke

from .types import QueryMode


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Key:
    value: str
    description: str
    name: str | None = None
    local_only: bool = False

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, str):
            return False
        return self.value == other


CANCEL_SELECTION = "KEY_ESCAPE"
CHANGE_DURATION_MODE = "T"
WRAP_QUERY = "v"
EXIT = "q"
HELP = "h"
SPACE = " "
COPY_TO_CLIPBOARD = "y"
PROCESS_CANCEL = "C"
PROCESS_KILL = "K"
PROCESS_FIRST = "KEY_HOME"
PROCESS_NEXT = "KEY_DOWN"
PROCESS_NEXTPAGE = "KEY_PGDOWN"
PROCESS_NEXT_VI = "j"
PROCESS_PREV = "KEY_UP"
PROCESS_PREV_VI = "k"
PROCESS_PREVPAGE = "KEY_PGUP"
PROCESS_LAST = "KEY_END"
PROCESS_PIN = Key(SPACE, "tag/untag current query", "Space")
REFRESH_DB_SIZE = "D"
REFRESH_TIME_INCREASE = "+"
REFRESH_TIME_DECREASE = "-"
SORTBY_READ = "r"
SORTBY_WRITE = "w"
SORTBY_MEM = "m"
SORTBY_TIME = "t"
SORTBY_CPU = "c"
HEADER_TOGGLE_SYSTEM = "s"
HEADER_TOGGLE_INSTANCE = "i"
HEADER_TOGGLE_WORKERS = "o"


def is_process_next(key: Keystroke) -> bool:
    if key.name == PROCESS_NEXT:
        return True
    elif key == PROCESS_NEXT_VI:
        return True
    return False


def is_process_prev(key: Keystroke) -> bool:
    if key.name == PROCESS_PREV:
        return True
    elif key == PROCESS_PREV_VI:
        return True
    return False


def is_process_nextpage(key: Keystroke) -> bool:
    return key.name == PROCESS_NEXTPAGE


def is_process_prevpage(key: Keystroke) -> bool:
    return key.name == PROCESS_PREVPAGE


def is_process_first(key: Keystroke) -> bool:
    return key.name == PROCESS_FIRST


def is_process_last(key: Keystroke) -> bool:
    return key.name == PROCESS_LAST


def is_toggle_header_system(key: Keystroke) -> bool:
    return key == HEADER_TOGGLE_SYSTEM


def is_toggle_header_instance(key: Keystroke) -> bool:
    return key == HEADER_TOGGLE_INSTANCE


def is_toggle_header_workers(key: Keystroke) -> bool:
    return key == HEADER_TOGGLE_WORKERS


EXIT_KEY = Key(EXIT, "quit")
PAUSE_KEY = Key(SPACE, "pause/unpause", "Space")

BINDINGS: list[Key] = [
    Key("Up/Down", "scroll process list"),
    PAUSE_KEY,
    Key(SORTBY_CPU, "sort by CPU% desc. (activities)", local_only=True),
    Key(SORTBY_MEM, "sort by MEM% desc. (activities)", local_only=True),
    Key(SORTBY_READ, "sort by READ/s desc. (activities)", local_only=True),
    Key(SORTBY_WRITE, "sort by WRITE/s desc. (activities)", local_only=True),
    Key(SORTBY_TIME, "sort by TIME+ desc. (activities)", local_only=True),
    Key(REFRESH_TIME_INCREASE, "increase refresh time (max:5s)"),
    Key(REFRESH_TIME_DECREASE, "decrease refresh time (min:0.5s)"),
    Key(WRAP_QUERY, "toggle query wrap"),
    Key(CHANGE_DURATION_MODE, "change duration mode"),
    Key(REFRESH_DB_SIZE, "force refresh database size"),
    Key("R", "force refresh"),
    Key(HEADER_TOGGLE_SYSTEM, "Display system information in header", local_only=True),
    Key(HEADER_TOGGLE_INSTANCE, "Display general instance information in header"),
    Key(HEADER_TOGGLE_WORKERS, "Display worker information in header"),
    EXIT_KEY,
]


def _sequence_by_int(v: int) -> tuple[str, str, int]:
    """
    >>> _sequence_by_int(11)
    ('F11', '11', 275)
    """
    assert 1 <= v <= 12, v
    return f"F{v}", str(v), getattr(curses, f"KEY_F{v}")


KEYS_BY_QUERYMODE = {
    QueryMode.activities: _sequence_by_int(1),
    QueryMode.waiting: _sequence_by_int(2),
    QueryMode.blocking: _sequence_by_int(3),
}
QUERYMODE_FROM_KEYS = {
    k: qm for qm, keys in KEYS_BY_QUERYMODE.items() for k in keys[1:]
}


MODES: list[Key] = [
    Key("/".join(KEYS_BY_QUERYMODE[qm][:-1]), qm.value) for qm in QueryMode
]
